#!/usr/bin/env python3

import argparse
import functools
import operator
import sys
import logging
import yaml
from http.client import HTTPConnection

from statograph import __version__
from statograph.datasources import Prometheus, Graphite
from statograph.statuspage import Statuspage
from statograph.uploader import MetricUploader

logger = logging.getLogger(__name__)


def _get(tree, branches, *, default=None):
    """Multi-level dict.get().  branches is an iterable of keys to traverse in tree."""
    try:
        return functools.reduce(operator.getitem, branches, tree)
    except KeyError:
        return default


def main():
    p = argparse.ArgumentParser(
        prog='statograph',
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    p.add_argument('-c', '--config', help='Path to your config file', required=True,
                   type=argparse.FileType('r'))
    p.add_argument('-v', '--verbose', action='count', default=0,
                   help='Verbosity (-v, -vv, etc).  Higher verbosities will log all HTTP traffic '
                        '(warning: at higher levels, this will log your auth secrets!)')
    p.add_argument('-V', '--version', action='version',
                   version='%(prog)s {}'.format(__version__))

    subparsers = p.add_subparsers(dest='command', metavar="COMMAND")

    subparsers.add_parser(
        'list_metrics',
        help="Display metadata about all metrics",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # TODO: this would be a nice feature to help users seed their configuration file,
    # but it is non-trivial to implement.
    # list_metrics.add_argument('-f', '--format', choices=['pretty', 'yaml'], default='pretty',
    #                           help="The output format")
    upload_metrics = subparsers.add_parser('upload_metrics',
                                           help="Upload new datapoints to metrics")
    upload_metrics.add_argument('metric_ids', nargs='*',
                                help='optional list of metric IDs to upload (empty == all)')

    erase = subparsers.add_parser('erase_metric_data',
                                  help="Erase historical data for a given metric")
    erase.add_argument('metric_id', help="The id string of the metric. Get this from list_metrics")

    # TODO: this would also be a nice feature, but won't be included in v0/v1.
    # subparsers.add_parser('new_incident', help="Open a new service outage")

    args = p.parse_args()

    loglevel = logging.WARNING
    if args.verbose >= 1:
        loglevel = logging.INFO
    if args.verbose >= 2:
        loglevel = logging.DEBUG
    if args.verbose >= 3:
        HTTPConnection.debuglevel = 1
        requests_log = logging.getLogger("urllib3")
        requests_log.setLevel(loglevel)
        requests_log.propagate = True

    logging.basicConfig(level=loglevel)

    config = yaml.safe_load(args.config.read())

    s = Statuspage(config['statuspage']['api_key'],
                   config['statuspage']['page_id'],
                   proxies=_get(config, ['proxies', 'statuspage']))

    if args.command == 'list_metrics':
        metrics = s.list_metrics()
        for metric in metrics:
            print(metric)
    elif args.command == 'erase_metric_data':
        s.reset_metric_data(args.metric_id)
    elif args.command == 'upload_metrics':
        everything_successful = True
        configured_metrics = {m['statuspage_id']: m for m in config['metrics']}
        to_upload = set(configured_metrics.keys())
        if args.metric_ids:
            unknown_metrics = set(args.metric_ids) - to_upload
            if unknown_metrics:
                logger.fatal("Asked to upload metrics without configuration data: %s",
                             unknown_metrics)
                raise KeyError(unknown_metrics)
            to_upload = set(args.metric_ids)
        graphite = Graphite(proxies=_get(config, ['proxies', 'graphite']))
        prometheus = Prometheus(proxies=_get(config, ['proxies', 'prometheus']))
        for metric_id in to_upload:
            try:
                if 'graphite' in configured_metrics[metric_id]:
                    datasource = graphite
                    base_url = configured_metrics[metric_id]['graphite']
                elif 'prometheus' in configured_metrics[metric_id]:
                    datasource = prometheus
                    base_url = configured_metrics[metric_id]['prometheus']
                else:
                    logger.warning("Unknown datasource for %s", metric_id)
                    continue
                uploader = MetricUploader(s.get_metric(metric_id))
                uploader.upload_one_batch(datasource, configured_metrics[metric_id]['query'],
                                          base_url, s)
            except:  # noqa: E722
                # No matter what exception happened, we want to log it and then continue onwards.
                # An erroneous query or broken datasource for one metric shouldn't prevent other
                # metrics from being updated.
                logger.error("Exception encountered when processing %s; continuing", metric_id,
                             exc_info=True)
                # That said, we should still report a problem to our executor.
                everything_successful = False

        return 0 if everything_successful else 1
        # TODO: export Prometheus textfile with most_recent_data stamps for each metric.
        # Then we can alert upon those timestamps being too stale.

    return 0


if __name__ == '__main__':
    sys.exit(main())
