"""Interfaces to timeseries data sources, like Prometheus and Graphite.

All the classes here implement a query_range function whose semantics are an
uneasy compromise between the Prometheus `query_range` HTTP API function and
Graphite's `render` HTTP API function.

See also https://prometheus.io/docs/prometheus/latest/querying/api/#range-queries
and https://graphite.readthedocs.io/en/stable/render_api.html"""

from typing import Dict, Iterable, Optional, Union

import requests

from statograph import Datapoint, set_user_agent


class PrometheusError(Exception):
    pass


class Prometheus:
    def __init__(self, proxies: Optional[Dict[str, str]] = None,
                 timeout_secs: Optional[float] = 20):
        """Construct a Prometheus query interface.

        If provided, proxies must be a requests-style dict from protocol to url: for example
        {'http': 'http://proxy:8080', 'https': 'http://proxy:8080'}"""
        self._session = requests.Session()
        self._timeout_secs = timeout_secs
        set_user_agent(self._session.headers)
        if proxies:
            self._session.proxies.update(proxies)

    def query_range(self, query: str, start: float, end: float,
                    base_url: str, step: float, max_datapoints: int) -> Iterable[Datapoint]:
        """Executes the given query, returning results between start and end every step seconds.

        start and end are UNIX epoch timestamps in seconds.  They must be at least
        step seconds apart, and should be no more than step*max_datapoints seconds apart.

        query must be a PromQL query string that returns a single scalar value.

        max_datapoints is ignored."""
        params: Dict[str, Union[int, float, str]] = {  # https://github.com/python/mypy/issues/3816
            'query': query, 'start': start, 'end': end, 'step': step,
        }
        r = self._session.get(base_url + '/api/v1/query_range',
                              timeout=self._timeout_secs,
                              params=params)
        r.raise_for_status()
        j = r.json()
        # See https://prometheus.io/docs/prometheus/latest/querying/api/#range-queries
        # for result format.
        if j['status'] == 'error':
            raise PrometheusError(j.get('error', 'Unknown error querying Prometheus'))
        if j['data']['resultType'] != 'matrix':
            raise PrometheusError('Did not get matrix result')
        metric_data = j['data']['result']
        if len(metric_data) != 1:
            raise PrometheusError(
                f"Query returned {len(metric_data)} timeseries instead of exactly 1")
        if len(metric_data[0]['metric']) != 0:
            raise PrometheusError('Query returned a single timeseries with labels still attached')
        for point in metric_data[0]['values']:
            yield Datapoint(point[0], float(point[1]))


class GraphiteError(Exception):
    pass


class Graphite:
    def __init__(self, proxies: Optional[Dict[str, str]] = None,
                 timeout_secs: Optional[float] = 20):
        """Construct a Graphite query interface.

        If provided, proxies must be a requests-style dict from protocol to url: for example
        {'http': 'http://proxy:8080', 'https': 'http://proxy:8080'}"""
        self._session = requests.Session()
        self._timeout_secs = timeout_secs
        set_user_agent(self._session.headers)
        if proxies:
            self._session.proxies.update(proxies)

    def query_range(self, query: str, start: float, end: float,
                    base_url: str, step: float, max_datapoints: int) -> Iterable[Datapoint]:
        """Executes the given query, returning up to max_datapoints results between start and end.

        start and end are UNIX epoch timestamps in seconds.

        query must be a Graphite query string.

        step is ignored, as Graphite has no notion of this in its query interface.  Assuming the
        number of datapoints between start and end is <= max_datapoints, Graphite will return
        the best temporal resolution available.  If more than max_datapoints are present in
        that interval, than datapoints will be consolidated down to max_datapoints.

        Make sure to worry about consolidateBy() in that case -- although with the default
        max_datapoints and step used by MetricsUploader, this isn't necessary in WMF's Graphite
        installation.  See also
        https://addshore.com/2018/09/grafana-graphite-and-maxdatapoints-confusion-for-totals/
        """
        params: Dict[str, Union[int, str]] = {  # https://github.com/python/mypy/issues/3816
            'target': query, 'from': int(start), 'until': int(end), 'format': 'json',
            'maxDataPoints': int(max_datapoints),
        }
        r = self._session.get(base_url + '/render',
                              timeout=self._timeout_secs,
                              params=params)
        r.raise_for_status()
        j = r.json()
        if len(j) != 1:
            raise GraphiteError(f"Query returned {len(j)} timeseries instead of exactly 1")
        # See https://graphite.readthedocs.io/en/stable/render_api.html#json for result format.
        for point in j[0]['datapoints']:
            if point[0]:
                yield Datapoint(timestamp=int(point[1]), value=point[0])
