"""Interface to Atlassian Statuspage's REST API.

A fairly trivial wrapper around several of their API methods.
See also https://developer.statuspage.io/"""

import logging
import time
from dataclasses import dataclass, asdict
from typing import Any, Dict, Collection, List, Optional

import requests
import dateutil.parser

from statograph import Datapoint, pretty_timestamp, set_user_agent

logger = logging.getLogger(__name__)


class StatuspageError(Exception):
    pass


@dataclass
class Metric:
    """A timeseries metric associated with a Statuspage page."""
    id: str
    name: str
    most_recent_data_at: Optional[float]  # UNIX epoch seconds

    def __str__(self) -> str:
        prefix = f"Metric '{self.name}' (id {self.id})"
        if self.most_recent_data_at:
            pretty_time = pretty_timestamp(self.most_recent_data_at)
            return (prefix + " with most recent data at "
                    f"{pretty_time} (@{self.most_recent_data_at})")
        else:
            return prefix + " with no recent data"


class Statuspage:
    """A simple wrapper around the Statuspage REST API.

    See also https://developer.statuspage.io/

    If provided, proxies must be a requests-style dict from protocol to url: for example
    {'http': 'http://proxy:8080', 'https': 'http://proxy:8080'}"""
    def __init__(self, api_key: str, page_id: str, *, proxies: Dict[str, str] = None,
                 timeout_secs: Optional[float] = 20):
        self._session = requests.Session()
        self._timeout_secs = timeout_secs
        set_user_agent(self._session.headers)
        self._session.headers['Authorization'] = f"OAuth {api_key}"
        if proxies:
            self._session.proxies.update(proxies)
        self._base_url = f"https://api.statuspage.io/v1/pages/{page_id}/"
        self._retry_after = 0.0  # UNIX epoch seconds after which it's ok to retry requests

    def _url(self, noun: str) -> str:
        return f"{self._base_url}{noun}"

    def _request(self, method: str, url: str, **kwargs) -> Any:
        """Execute a request against self._session, respecting ratelimiting and raising if
        the response isn't okay.  Retry on ratelimiting (after sleeping) but not on other error.

        The kwargs interface of this function follows the interface of Session.request().

        Statuspage API calls can return either a dict or a list."""
        # Respect any previous ratelimiting
        time.sleep(max(0, self._retry_after - time.time()))
        resp = self._session.request(method, url, timeout=self._timeout_secs, **kwargs)
        if resp.status_code == 429:
            self._retry_after = time.time() + float(resp.headers.get('retry-after', 2))
            logger.warn('Got ratelimited by Statuspage; retrying at %s seconds for %s',
                        self._retry_after, url)
            return self._request(method, url, timeout=self._timeout_secs, **kwargs)
        else:
            try:
                resp.raise_for_status()
            except requests.HTTPError:
                logger.error('Error response from Statuspage: %s', resp.content.decode())
                raise
        j = resp.json()
        logger.debug('resp for %s to %s json: %s', method, url, j)
        return j

    def _metric_parse_json(self, json: Dict[str, Any]) -> Metric:
        """Parse API-returned JSON into a Metric object."""
        most_recent = None
        if json.get('most_recent_data_at', None):
            most_recent = dateutil.parser.isoparse(json['most_recent_data_at']).timestamp()
        return Metric(id=json['id'], name=json['name'], most_recent_data_at=most_recent)

    def list_metrics(self) -> List[Metric]:
        """Fetch metadata for all metrics associated with our page."""
        all_metrics = self._request('GET', self._url('metrics'))
        return [self._metric_parse_json(metric) for metric in all_metrics]

    def metrics_dict(self) -> Dict[str, Metric]:
        """Fetch metadata for all metrics associated with our page, as a dict from id to Metric."""
        return {m.id: m for m in self.list_metrics()}

    def get_metric(self, metric_id: str) -> Metric:
        """Fetch metadata for a given metric_id."""
        return self._metric_parse_json(self._request('GET', self._url(f"metrics/{metric_id}")))

    def reset_metric_data(self, metric_id: str) -> Metric:
        """Purge all historical data for a given metric_id."""
        return self._metric_parse_json(
            self._request('DELETE', self._url(f"metrics/{metric_id}/data")))

    def add_datapoints(self, data: Dict[str, Collection[Datapoint]]) -> Dict[str, Any]:
        """Given a dictionary from metric_id to a collection of Datapoints, upload them.

        Will raise an HTTPError if too many datapoints are provided or if format is invalid."""
        body = {}
        body['data'] = {metric_id: [asdict(t) for t in data[metric_id]]
                        for (metric_id, t) in data.items()}
        logger.info("Uploading %s datapoints", sum(len(v) for v in data.values()))
        return self._request('POST', self._url("metrics/data"), json=body)
