"""Connects timeseries datasources with Statuspage.

It is possible to upload metric data to Statuspage in mostly-arbitrary
intervals of time.  However, the only visibility into what has been uploaded
that is accessible via the API is a single 'most_recent_data' timestamp.
Therefore, we always upload in FIFO order.  This means there is no need to
maintain any state on the client side -- the API can be the only source of truth.

Taking the most_recent_data timestamp as our start time, we then upload as much
as is needed while still fitting in a single Statuspage API call.

In the usual case, a data upload will be small (just a few of the most recent
data points).  In the case of a new metric, or a metric whose history was purged,
the backlog will still naturally fill assuming the uploader is invoked in a
loop, such as via a systemd timer.

We make some minimal concessions to clock skew, but hosts should be synchronized to
within a few minutes of each other.
"""

import logging
import time

from statograph import pretty_timestamp
from statograph.statuspage import Metric, Statuspage


logger = logging.getLogger(__name__)


class MetricUploader:
    """Glue between Datasources and Statuspage.  Uploads in FIFO order."""
    def __init__(self, metric: Metric, *, now=time.time):
        """Given a Metric, this object can compute time bounds on what data to upload and also
        perform the upload.  now must be a function like time.time() that returns
        UNIX epoch seconds -- exposed for testability."""
        self.metric = metric
        self.now = now()

    # Statuspage throws HTTP 4xxs if you provide data further back in time than 28 days.
    # This will fill ~99.9% of history while also minimizing risk due to any clock skew.
    MAX_DAYS_BACKWARDS = 27+(23/24)
    # 1500 was empirically determined to be close to the max in one API call.
    # Unfortunately the actual maximum is not documented.
    # Because of [fence, post] issues around timestamps in Prometheus, we may
    # upload one additional data point on top of the max when doing a backfill.
    # At 1500 this is okay.
    MAX_POINTS_PER_REQUEST = 1500
    # Statuspage doesn't really support more than one-minute granularity on metrics --
    # they'll accept data at shorter intervals but always seem to downsample for display.
    # 60s is also convenient in that this is the shortest interval available in our Graphite,
    # and a multiple of the shortest window available in our Prometheus (30 seconds).
    STEP_SECONDS = 60

    def start_timestamp(self) -> float:
        """At what timestamp should our next upload begin?

        Exposed for testability."""
        most_recent_data = self.metric.most_recent_data_at or 0
        earliest_valid = self.now - self.MAX_DAYS_BACKWARDS*86400
        return max(most_recent_data, earliest_valid)

    def end_timestamp(self) -> float:
        """At what timestamp should our next upload end?

        Exposed for testability."""
        start = self.start_timestamp()
        largest_possible_upload = start + self.MAX_POINTS_PER_REQUEST*self.STEP_SECONDS
        return min(largest_possible_upload, self.now)

    def upload_one_batch(self, datasource, query: str, base_url: str, statuspage: Statuspage):
        """Given a timeseries datasource, a query for that datasource, and the base URL of that
        data source, perform a query and upload the results to Statuspage."""
        start = self.start_timestamp()
        end = self.end_timestamp()
        if start >= end - self.STEP_SECONDS:
            logger.info("Too soon to upload for metric '%s' (most_recent_data at %s, now is %s)",
                        self.metric.name, pretty_timestamp(self.metric.most_recent_data_at),
                        pretty_timestamp(self.now))
            return
        logger.info("Querying data for metric name '%s' from %s to %s", self.metric.name,
                    pretty_timestamp(start), pretty_timestamp(end))
        data = list(datasource.query_range(query, start, end, base_url,
                                           self.STEP_SECONDS, self.MAX_POINTS_PER_REQUEST))
        if data:
            statuspage.add_datapoints({self.metric.id: data})
